# Generated by roxygen2: do not edit by hand

S3method(as.promise,Future)
S3method(as.promise,default)
S3method(as.promise,promise)
S3method(format,promise)
S3method(is.promising,Future)
S3method(is.promising,default)
S3method(is.promising,promise)
S3method(print,promise)
export("%...!%")
export("%...>%")
export("%...T!%")
export("%...T>%")
export("%>%")
export("%T>%")
export(as.promise)
export(catch)
export(finally)
export(future_promise)
export(future_promise_queue)
export(hybrid_then)
export(is.promise)
export(is.promising)
export(local_ospan_promise_domain)
export(local_otel_promise_domain)
export(new_promise_domain)
export(promise)
export(promise_all)
export(promise_map)
export(promise_race)
export(promise_reduce)
export(promise_reject)
export(promise_resolve)
export(then)
export(with_ospan_async)
export(with_ospan_promise_domain)
export(with_otel_promise_domain)
export(with_otel_span)
export(with_promise_domain)
importFrom(R6,R6Class)
importFrom(later,later)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(magrittr,"%T>%")
importFrom(otel,end_span)
importFrom(otel,get_active_span)
importFrom(otel,is_tracing_enabled)
importFrom(otel,start_span)
importFrom(otel,with_active_span)
