# promises <a href="https://rstudio.github.io/promises/"><img src="man/figures/logo.svg" align="right" height="139" alt="promises website" /></a>

<!-- badges: start -->
[![R-CMD-check](https://github.com/rstudio/promises/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/promises/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

A promise library for R. https://rstudio.github.io/promises/

## Installation

```r
install.packages("promises")
```

To use promises with Shiny, make sure you have Shiny v1.1.0 or later.

## License

MIT
