# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

asyncFib <- function(resolve, reject, x) {
    invisible(.Call('_promises_asyncFib', PACKAGE = 'promises', resolve, reject, x))
}

